MODULE CodeTables
CONTAINS

!-----------------------------------------------------------------------
!  codetables.f90 - Module of special BUFR Code Table look-up routines
!
!  CALL CT001033   ( Code, Name, ICAO )  : Originating Centre look-up
!  CALL CT001034   ( Code, Name )        : Originating Sub-centre look-up
!  CALL TableA     ( Code, Name )        : Table A data type look-up
!  CALL LoadCTable ( Table, Default, &
!                    Entry[, UC_Entry] ) : Load a code table
!  CALL CTUpper    ( string )            : Convert lowercase to UPPERCASE
!-----------------------------------------------------------------------
!
SUBROUTINE CT001033 ( Code, & ! (inout)
                      Name, & ! (inout)
                      ICAO )  ! (inout)
!****s* BUFR/CT001033 *
!-----------------------------------------------------------------------
!
! NAME
!   CT001033     (codetables.f90)
!
! SYNOPSIS
!   Look up the Originating/Generating Centre in BUFR Code Table 001033
!   and its associated ICAO code
!
!    USE codetables
!    INTEGER            :: CODE
!    CHARACTER (LEN=4)  :: ICAO
!    CHARACTER (LEN=50) :: NAME
!    CODE=74
!    CALL CT001033(CODE,NAME,ICAO) ! look up name & ICAO for given code
!    CODE=-1
!    ICAO='EGRR'
!    CALL CT001033(CODE,NAME,ICAO) ! look up code & name for given ICAO
!    CODE=-1
!    NAME='UK MET'
!    CALL CT001033(CODE,NAME,ICAO) ! look up code & ICAO for given name
!
! ARGUMENTS
!   Code  int  Originating/Generating Centre code value (0-255)
!   Name  chr  Originating/Generating Centre name for code value
!   ICAO  chr  ICAO 4-character code associated with Orig/Gen Centre
!
! FILES
!   CT001033  - in path defined by BUFR_LIBRARY
!
! ENVIRONMENT VARIABLES
!   BUFR_LIBRARY - path to run-time BUFR tables
!
! CALLS
!   BUFRPATH
!
! DESCRIPTION
!   BUFR Code Table 001033 (Common Code Table C-1) is a list of codes
!   identifying the originating or generating centre. Most centres
!   have an associated 4-chr ICAO code used e.g. for WMO abbreviated
!   routing headers.
!   Given an Originating/Generating Centre code value (0-254), this 
!   subroutine returns the corresponding centre name from Code Table 
!   001033 plus its associated ICAO code (where known). 
!   Codes not yet defined are returned as 'Reserved'; Codes out 
!   of range (>255) are returned 'INVALID CODE VALUE'. Where ICAO codes
!   are not known, or the code value is invalid, the ICAO string will be 
!   returned as 'UNKN'.
!   If the given code value is <0, then the text argument(s) are used to 
!   look up any matching string(s) and if found, returns the corresponding
!   code value (0-254), full matching name and ICAO strings. If no match 
!   is found, the code value is returned as 255, the name set to 'MISSING'
!   ICAO set to 'UNKN'. Matching is not case sensitive.
!   The Code Table file is loaded on the first call, but if the file
!   cannot be loaded, the returned name is 'TABLE 001033 NOT LOADED',
!   except for code 255 ('MISSING'). Any missing entries in the file
!   are assumed to be 'Reserved' with the ICAO being 'UNKN'.
!   The file name is 'CT001033' and is located via environment variable 
!   'BUFR_LIBRARY'.
!
! MODIFICATION HISTORY
!   Version  Date         Change                       By
!  -----------------------------------------------------------------
!   1.00     25-Jul-2006  First release                D. Offiler
!
! COPYRIGHT
!   Language:           Fortran 90
!   Software Standards: GTDP 8
!
! (c) CROWN COPYRIGHT 2006, Met Office, All Rights Reserved.
!
! Use, duplication or disclosure of this code is subject to the
! restrictions as set forth in the contract.
!
!                Met Office, FitzRoy Road,
!                Exeter, Devon, EX1 3PB, UK
!
! If no contract has been raised with this copy of the code, the use,
! duplication and disclosure of it is strictly prohibited. Permission
! to do so must first be obtained in writing, from the Head of Satellite
! Applications at the above address.
!
!-------------------------------------------------------
!****
!  
  IMPLICIT NONE

! Fixed values

  CHARACTER (LEN=*), PARAMETER :: Tdsn  = "CT001033"  ! Table file name
  INTEGER,           PARAMETER :: Tunit = 8           ! Table stream unit
  
! Argument list parameters

  INTEGER,           INTENT(INOUT) :: Code
  CHARACTER (LEN=*), INTENT(INOUT) :: Name
  CHARACTER (LEN=*), INTENT(INOUT) :: ICAO

! Local parameters

  CHARACTER (LEN=4)   :: UC_ICAO
  CHARACTER (LEN=60)  :: Entry(0:255) = " "
  CHARACTER (LEN=60)  :: UC_Entry(0:255), UC_Name
  LOGICAL             :: Loaded = .FALSE.
  
  SAVE Entry
  SAVE UC_Entry
  SAVE Loaded

! For first entry, load Code Table 001033

  IF ( .NOT. Loaded ) THEN
    Entry(0:254) = "UNKN TABLE 001033 NOT LOADED"
    Entry(255)   = "UNKN MISSING"
    CALL LoadCTable ( Tdsn, "UNKN Reserved", Entry, UC_Entry )
  END IF

! Decode given code value to name & ICAO strings...

  IF ( Code >= 0 ) THEN

    IF ( Code <= 255 ) THEN
      ICAO = Entry(Code)(1:4)
      IF ( ICAO == " " ) ICAO = "UNKN"
      Name = Entry(Code)(6:)
    ELSE
      ICAO = "UNKN"
      Name = "INVALID CODE VALUE"
    END IF
    
! ...or look up code value for (partial) matching text - 
! as name or ICAO. Then decode to full strings as above.
! All matching is done in UPPERCASE

  ELSE
    IF ( ICAO == " " ) THEN
      UC_ICAO = "?"
    ELSE
      UC_ICAO = ICAO
      CALL CTUpper ( UC_ICAO )
    END IF 
    IF ( Name == " " ) THEN
      UC_Name = "?"
    ELSE
      UC_Name = Name
      CALL CTUpper ( UC_Name )
    END IF
    
    Code = 0    
    DO
      IF ( Code == 255                                      .OR. &
           INDEX ( UC_Entry(Code)(1:4), TRIM(UC_ICAO) ) > 0 .OR. &                                         
           INDEX ( UC_Entry(Code)(6:),  TRIM(UC_Name) ) > 0 ) EXIT
      Code = Code + 1
    END DO

    IF ( Code <= 255 ) THEN
      ICAO = Entry(Code)(1:4)
      IF ( ICAO == " " ) ICAO = "UNKN"
      Name = Entry(Code)(6:)
    ELSE
      ICAO = "UNKN"
      Name = "INVALID CODE VALUE"
    END IF

  END IF

END SUBROUTINE CT001033
!---------------------------------------------------------------
SUBROUTINE CT001034 ( Code, & ! (inout)
                      Name )  ! (inout)
!****s* BUFR/CT001034 *
!-----------------------------------------------------------------------
!
! NAME
!   CT001034     (codetables.f90)
!
! SYNOPSIS
!   Look up the Originating Sub-Centre in BUFR Code Table 001034
!
!    USE codetables
!    INTEGER            :: CODE
!    CHARACTER (LEN=20) :: NAME
!    CODE=1
!    CALL CT001034(CODE,NAME) ! look up sub-centre name for given code
!    CODE=-1
!    NAME='Kiruna'
!    CALL CT001033(CODE,NAME) ! look up sub-centre code for given name
!
! ARGUMENTS
!   Code  int  Originating Sub-Centre code value (0-255)
!   Name  chr  Originating Sub-Centre text for code value
!
! FILES
!   CT001034  - in path defined by BUFR_LIBRARY
!
! ENVIRONMENT VARIABLES
!   BUFR_LIBRARY - path to run-time BUFR tables
!
! CALLS
!   BUFRPATH
!
! DESCRIPTION
!   BUFR Code Table 001034 is a locally-defined list of Sub-Centres
!   valid for an associated Originating Centre (001031 or 001033).
!   Given an Originating Sub-Centre code value (0-255), this subroutine
!   returns the corresponding csub-centre name from Code Table 001034. 
!   Codes not yet defined are returned as 'Reserved'; Codes out 
!   of range are returned 'INVALID CODE VALUE'.
!   If the given code value is <0, then the text argument is used to 
!   look up any matching name string and if found, returns the 
!   corresponding code value (0-254). If no match is found, the value
!   is returned as 255 (equivalent to 'MISSING').
!   The Code Table file is loaded on the first call, but if the  file
!   cannot be loaded, the return is 'TABLE 001034 NOT LOADED',
!   except for codes 0 ('NONE') and 255 ('MISSING').
!   The file name is 'CT001034' and is located via environment variable 
!   'BUFR_LIBRARY'.
!
! MODIFICATION HISTORY
!   Version  Date         Change                       By
!  -----------------------------------------------------------------
!   1.00     08-Apr-2004  First release                D. Offiler
!   1.01     09-Jun-2004  If more than one entry per code value,
!                         retain only the first.       D. Offiler
!   1.02     26-Jul-2006  Re-written to be compatible with
!                         new CT001033/LoadCtable and slightly 
!                         modified file format.        D. Offiler
!
! COPYRIGHT
!   Language:           Fortran 90
!   Software Standards: GTDP 8
!
! (c) CROWN COPYRIGHT 2006, Met Office, All Rights Reserved.
!
! Use, duplication or disclosure of this code is subject to the
! restrictions as set forth in the contract.
!
!                Met Office, FitzRoy Road,
!                Exeter, Devon, EX1 3PB, UK
!
! If no contract has been raised with this copy of the code, the use,
! duplication and disclosure of it is strictly prohibited. Permission
! to do so must first be obtained in writing, from the Head of Satellite
! Applications at the above address.
!
!-------------------------------------------------------
!****
!  
  IMPLICIT NONE

! Fixed Codes

  CHARACTER (LEN=*), PARAMETER :: Tdsn  = "CT001034"  ! Table file name
  INTEGER,           PARAMETER :: Tunit = 8           ! Table stream unit
  
! Argument list parameters

  INTEGER,           INTENT(INOUT) :: Code
  CHARACTER (LEN=*), INTENT(INOUT) :: Name

! Local parameters

  CHARACTER (LEN=60)  :: Entry(0:255) = " "
  CHARACTER (LEN=60)  :: UC_Entry(0:255), UC_Name
  LOGICAL             :: Loaded = .FALSE.
  
  SAVE Entry, UC_Entry
  SAVE Loaded

! For first entry, load Code Table 001034 categories

  IF ( .NOT. Loaded ) THEN
    Entry(0)     = "NONE"
    Entry(1:254) = "TABLE 001034 NOT LOADED"
    Entry(255)   = "MISSING"
    CALL LoadCTable ( Tdsn, "Reserved", Entry, UC_Entry )
  END IF

! Decode given code Code to textual meaning...

  IF ( Code >= 0 ) THEN
    IF ( Code <= 255 ) THEN
      Name = Entry(Code)
    ELSE
      Name= "INVALID CODE VALUE"
    END IF
    
! ...or look up code value for (partial) matching text.
! Then decode to full name as above.
! All matching is done in UPPERCASE

  ELSE
    IF ( Name == " " ) THEN
      UC_Name = "?"
    ELSE
      UC_Name = Name
      CALL CTUpper ( UC_Name )
    END IF
    
    Code = 0
    DO
      IF ( Code == 255 .OR. &
           INDEX ( UC_Entry(Code), TRIM(UC_Name) ) > 0 ) EXIT
      Code = Code + 1
    END DO
    Name = Entry(Code)

  END IF

  
END SUBROUTINE CT001034
!-----------------------------------------------------------------------

SUBROUTINE TableA ( Code, & ! (in)
                    Name )  ! (out)
!****s* BUFR/CTUpper *
!-----------------------------------------------------------------------
!
! NAME
!   TableA     (codetables.f90)
!
! SYNOPSIS
!   Look up the data type in BUFR Table A
!
!    USE codetables
!    INTEGER            :: CODE
!    CHARACTER (LEN=20) :: NAME
!    CODE=12
!    CALL TABLEA(CODE,NAME) ! look up data type for given code
!
! ARGUMENTS        
!   Code  int  (in)  Table A code Code (0-255)
!   Name  chr  (out) Table A text for code value
!
! Calls:
!   BUFRPATH
!
! FILES
!   TABLEA  - in path defined by BUFR_LIBRARY
!
! ENVIRONMENT VARIABLES
!   BUFR_LIBRARY - path to run-time BUFR tables
!
! CALLS
!   BUFRPATH
!
! DESCRIPTION
!   Given a Table A code value (0-255), returns the corresponding
!   data category description from Table A. Codes not yet defined 
!   are returned as 'Reserved'; Codes out of range are returned 
!   'INVALID CODE VALUE'.  The BUFR Table A file is loaded on the
!   first call, but if the file cannot be loaded, the return is
!   'TABLE A NOT LOADED'. The file name is 'TABLEA' and is located
!   via environment variable 'BUFR_LIBRARY'.
!
! MODIFICATION HISTORY
!   Version  Date         Change                       By
!  -----------------------------------------------------------------
!   1.00     13-Dec-2001  First f90 release            D. Offiler
!
! COPYRIGHT
!   Language:           Fortran 90
!   Software Standards: GTDP 8
!
! (c) CROWN COPYRIGHT 2006, Met Office, All Rights Reserved.
!
! Use, duplication or disclosure of this code is subject to the
! restrictions as set forth in the contract.
!
!                Met Office, FitzRoy Road,
!                Exeter, Devon, EX1 3PB, UK
!
! If no contract has been raised with this copy of the code, the use,
! duplication and disclosure of it is strictly prohibited. Permission
! to do so must first be obtained in writing, from the Head of Satellite
! Applications at the above address.
!
!-------------------------------------------------------
!****
!  
  IMPLICIT NONE

! Fixed values

  CHARACTER (LEN=*), PARAMETER :: Tdsn  = "TABLEA"  ! Table A file name
  INTEGER,           PARAMETER :: Tunit = 8         ! Table A stream unit
  
! Argument list parameters

  INTEGER,           INTENT(IN)  :: Code
  CHARACTER (LEN=*), INTENT(OUT) :: Name

! Local parameters

  CHARACTER (LEN=60)  :: Entry(0:255) = " "
  LOGICAL             :: Loaded = .FALSE.
  
  SAVE Entry
  SAVE Loaded

! For first entry, load Table A

  IF ( .NOT. Loaded ) THEN
    CALL LoadCTable ( Tdsn, "Reserved", Entry )
    Loaded = .TRUE.
  END IF

! Decode given Table A code value to textual meaning

  IF ( Code >= 0 .AND. Code <= 255 ) THEN
    Name = Entry(Code)
  ELSE
    Name= "INVALID CODE VALUE"
  END IF

END SUBROUTINE TableA
!-------------------------------------------------------

SUBROUTINE LoadCTable ( Table,     & ! (in)
                        Default,   & ! (in)
                        Entry,     & ! (inout)
                        UC_Entry )   ! (out; optional)

!****s* BUFR/LoadCTable *
!-----------------------------------------------------------------------
!
! NAME
!   LoadCTable     (codetables.f90)
!
! SYNOPSIS
!   Subroutine to load a special BUFR code table
!
!    USE CodeTables
!    CHARACTER (LEN=20) :: Table           = 'TABLEA'
!    CHARACTER (LEN=20) :: Default         = 'Reserved'
!    CHARACTER (Len=60) :: Entry(0:255)    = " "
!    CHARACTER (Len=60) :: UC_Entry(0:255) [optional]
!    Entry(255)="MISSING"
!    CALL LoadCTable ( Table, Default, Entry )
!  or 
!    CALL LoadCTable ( Table, Default, Entry, UC_Entry )
!
! ARGUMENTS
!   Table    (in)     chr  Name of code table file
!   Default  (in)     chr  Default string (usually 'Reserved')
!   Entry    (inout)  chr  (0:255) array of code table entries
!   UC_entry (out)    chr  (0:255) array of code table entries
!                          in UPPERCASE. Optional argument.
!
! FILES
!   'Table' string - in path defined by BUFR_LIBRARY. Uses stream 8.
!
! ENVIRONMENT VARIABLES
!   BUFR_LIBRARY - path to run-time BUFR tables
!
! CALLS
!   BUFRPATH
!
! DESCRIPTION
!   Loads a special BUFR code table from a file. The file is found
!   in the path defined by environment variable 'BUFR_LIBRARY'.
!   The file may contain up to 256 entries formatted thus:
!    nnnn text
!   where 'nnn' starts in column 1 and is either:
!    - a single 3-digit number 000-255 or
!    - a pair of 3-digit numbers separated by a dash ('-')
!      with the second number not smaller than the first, representing
!      a range of entries having identical text
!   and 'text' is free-format text (mixed case) up to 60 characters.
!   Codes may appear in any order, but numerically ascending is 
!   advised! Text for any duplicate codes will overwrite any earlier ones.
!   Any line not starting with a valid 3-digit number is ignored.
!   If the file does not contain entries for all possible codes 0-255
!   then the 'Default' string (usually 'Reserved') is used to pad the 
!   'Entry' array.
!   Optionally, the strings in 'Entry' can be converted to all
!   uppercase (in optional argument 'UC_Entry') for matching purposes.
!   Before calling this subroutine, the 'Entry' array must be initialised 
!   to at least blanks; any blank entry is first initialised to the 
!   string: 'TABLE <table> NOT LOADED' so that should the file not be 
!   found or not read, this string will indicate that fact. Any entry
!   pre-initialised to a non-blank string before calling this subroutine 
!   will not be re-initialised, so that known codes can be set even if 
!   the file cannot be loaded. E.g. Entry(255) may be 'Missing' or 
!   'Not used'. A valid file entry will always overwrite a pre-initialised
!   entry.
!
! MODIFICATION HISTORY
!
! Version Date          Comment
! ------- ----          -------
! 1.0     26-Jul-2006   First release.                      D. Offiler
!
! COPYRIGHT
!   Language:           Fortran 90
!   Software Standards: GTDP 8
!
! (c) CROWN COPYRIGHT 2006, Met Office, All Rights Reserved.
!
! Use, duplication or disclosure of this code is subject to the
! restrictions as set forth in the contract.
!
!                Met Office, FitzRoy Road,
!                Exeter, Devon, EX1 3PB, UK
!
! If no contract has been raised with this copy of the code, the use,
! duplication and disclosure of it is strictly prohibited. Permission
! to do so must first be obtained in writing, from the Head of Satellite
! Applications at the above address.
!
  IMPLICIT NONE

! Fixed values

  INTEGER, PARAMETER :: Tunit  = 8         ! Table stream unit

! Argument list parameters

  CHARACTER (LEN=*), INTENT(IN)            :: Table
  CHARACTER (LEN=*), INTENT(IN)            :: Default
  CHARACTER (LEN=*), INTENT(OUT)           :: Entry(0:255)
  CHARACTER (LEN=*), INTENT(OUT), OPTIONAL :: UC_Entry(0:255)

! Local variables

  CHARACTER (LEN=64)  :: Line
  CHARACTER (LEN=100) :: Dir
  INTEGER             :: ierr
  INTEGER             :: i1, i2
  INTEGER             :: LenDir

! Initialise
  
  WHERE ( Entry == " " ) 
    Entry = "TABLE "//TRIM(Table)//" NOT LOADED"
  END WHERE
  
  CALL BUFRPATH ( Dir, LenDir )

! Open Table file

  OPEN ( UNIT=Tunit,                &
         FILE=Dir(1:LenDir)//Table, &
       STATUS="OLD",                &
       IOSTAT=ierr )

! Read Table entries

  DO WHILE ( ierr == 0 )
    READ ( UNIT=Tunit,  &
            FMT="(A)", &
         IOSTAT=ierr ) Line

    IF ( ierr /= 0 ) THEN
      EXIT
    ELSE IF ( Line(4:4) == "-" ) THEN
      READ ( Line(1:7), "(I3,1X,I3)" ) i1, i2
      Entry(i1:i2) = Line(9:)
    ELSE IF ( LGE(Line(1:3),"000") .AND. &
              LLE(Line(1:3),"255") ) THEN
      READ ( Line(1:3), "(I3)" ) i1
      Entry(i1) = Line(5:)
    ELSE
      CYCLE
    END IF
  END DO
  CLOSE ( UNIT=Tunit )

! If table successfully loaded, fill in any undefined entries
! with a default string

  IF ( ierr <= 0 ) THEN
    DO i1 = 0, 254
      IF ( INDEX(Entry(i1), "NOT LOADED") > 0 ) &
        Entry(i1) = Default
    END DO
  END IF

! Create an (optional) uppercase copy for matching

  IF ( PRESENT(UC_Entry) ) THEN
    DO i1 = 0, 255
      UC_Entry(i1) = Entry(i1)
      CALL CTUpper ( UC_Entry(i1) )
    END DO
  END IF

END SUBROUTINE LoadCTable
!-----------------------------------------------------------------------

SUBROUTINE CTUpper ( string )

!****s* BUFR/CTUpper *
!-----------------------------------------------------------------------
!
! NAME
!   CTUpper     (codetables.f90)
!
! SYNOPSIS
!   Subroutine to convert alphabetic characters from lower to upper case
!
!    USE codetables
!    CHARACTER (LEN=20) :: string = " ABcdE!"
!    CALL CTUpper ( string )
!    [string = " ABCDE!"]
!
! ARGUMENTS
!   string  (inout)  chr  string to be converted
!
! DESCRIPTION
!   Scans through input string for lowercase characters [a-z]
!   and replaces them with uppercase equivalents [A-Z]
!   This version taken from ToLower in chars.f90 in Genlib package.
!
! MODIFICATION HISTORY
!
! Version Date          Comment
! ------- ----          -------
! 1.0     26-Jul-2006   First release.                      D. Offiler
!
! COPYRIGHT
!   Language:           Fortran 90
!   Software Standards: GTDP 8
!
! (c) CROWN COPYRIGHT 2006, Met Office, All Rights Reserved.
!
! Use, duplication or disclosure of this code is subject to the
! restrictions as set forth in the contract.
!
!                Met Office, FitzRoy Road,
!                Exeter, Devon, EX1 3PB, UK
!
! If no contract has been raised with this copy of the code, the use,
! duplication and disclosure of it is strictly prohibited. Permission
! to do so must first be obtained in writing, from the Head of Satellite
! Applications at the above address.
!
!-------------------------------------------------------
!****

  IMPLICIT NONE
  
  CHARACTER (LEN=*),  INTENT(INOUT) :: string
  
  CHARACTER (LEN=26), PARAMETER :: UPPER="ABCDEFGHIJKLMNOPQRSTUVWXYZ"
  CHARACTER (LEN=26), PARAMETER :: lower="abcdefghijklmnopqrstuvwxyz"
  INTEGER :: i, j

  DO i = 1, LEN_TRIM(string)
    j = INDEX ( lower, string(i:i) )
    IF ( j > 0 ) string(i:i) = UPPER(j:j)
  END DO

END SUBROUTINE CTUpper
!-------------------------------------------------------

END MODULE CodeTables
