!--------------------------------------------------------------------
!
! Code description:
!   Language:           Fortran 90
!   Software Standards: GTDP 8
!
! (c) CROWN COPYRIGHT 2008, METEOROLOGICAL OFFICE, All Rights Reserved.
!
! Use, duplication or disclosure of this code is subject to the
! restrictions as set forth in the contract.
!
!                The Met. Office, London Road,
!                BRACKNELL, Berkshire, UK, RG12 2SZ
!
! If no contract has been raised with this copy of the code, the use,
! duplication and disclosure of it is strictly prohibited. Permission
! to do so must first be obtained in writing, from the Head of Satellite
! Applications at the above address.
!
! History:
!   Version  Date      Change                            By
!  -----------------------------------------------------------------
!   1.00     Dec 2001  First f90 release                 D. Offiler
!   1.01     May 2008  Define max. message length parameter here;
!                      increase value from 25K to 500K and increase
!                      max. no. descr. from 20K to 200K to cope with 
!                      new GTS message length limit of 500Kb wef 
!                      5 Nov 2007.                       D. Offiler
!
!--------------------------------------------------------------------
MODULE DecBUFRmod

! Fixed values

  CHARACTER (LEN=*), PARAMETER :: Version = "V1.09 - 27-Aug-2008"

  INTEGER, PARAMETER :: nb      =   500000  ! Max length of BUFR message
  INTEGER, PARAMETER :: nd      =   200000  ! Max. no. of descriptors
  INTEGER, PARAMETER :: no      =      500  ! Max. no. of observations
  INTEGER, PARAMETER :: nv      =    nd*no  ! Max. no. of data values
  INTEGER, PARAMETER :: Missing = -9999999  ! Missing data indicator
  INTEGER, PARAMETER :: Ounit   =        8  ! Output data stream unit no.
  
  INTEGER, PARAMETER :: ErrOK   =        0  ! No  error
  INTEGER, PARAMETER :: ErrIO   =        1  ! I/O error
  INTEGER, PARAMETER :: ErrMem  =        2  ! Memory allocation error
  
END MODULE DecBUFRMod
!------------------------------------------------------------------------
PROGRAM DecBUFR
!
! Purpose:
!   Generic BUFR decoder
!
! Calls:
!   DecBUFRoptions
!   AnalBUFR
!   OutputVals
!   BUFRREAD
!   DEBUFR
!
! Modules:
!   DecBUFRmod  - fixed parameter definitions
!
! Dependencies:
!   MetDB BUFR kernel package (libbufr.a)
!   BUFR tables (TABLEA, TABLEB, TABLED, CODEFIG, CT001033, CT001034)
!        - found via environment variable 'BUFR_LIBRARY'
!
! Description:
!   Provides a simple generic BUFR decoder user interface (decodes
!   arbitrary BUFR-encoded data). BUFR tables are found via the 
!   environment variable 'BUFR_LIBRARY'.
!   Invoke as:
!     > decbufr bufr_file [-d[h]]   [-o output_file]
!                         [-f first | -s skip] 
!                         [-l last  | -n number]
!                         [-h | ?]
!    where:
!      bufr_file   is the input file containing BUFR message(s)
!      -d          writes an interpreted diagnostic dump to stdout
!      -dh         writes an interpreted headers dump only to stdout
!      output_file is a (formatted text) data file
!      first       is the first message to decode (ie skip first-1)
!      skip        is the number of messages to skip (ie first=skip+1)
 !      last        is the last message to decode (ie number=first-last+1)
!      number      is the no. of messages to decode from first.
!    Arguments can be in any order and any space(s) between a
!    switch and its value is optional. If both -f & -s or both 
!    -l & -n options are present, only the second is taken. 
!    If the bufr_file is not specified, or if '?' or '-h', options
!    are present, a usage reminder only is printed and no decoding
!    is done. The output file is optional, and if not specified,
!    is not created. e.g.:
!      > decbufr bufr.gts -o bufr.out
!    will decode all messages from file bufr.gts and output a
!    formatted data file bufr.out; no dump data is written.
!    Limitations:
!    a) up to 5000 parameters per observation
!    b) up to 4000 observations per message
!    c) does not output data from section 2 if present.
!    d) the BUFR details for the data must be present in at
!       least Tables B & D for the decode to be successful.
!
! History:
!   Version  Date      Change                            By
!  -----------------------------------------------------------------
!   1.00     Dec 2001  First f90 release                 D. Offiler
!   1.01     Jun 2002  a) Output BUFR file read errors 
!                         in plain text instead of just 
!                         an error code.  
!                      b) Signal exit error code.  
!                      c) Remove EBCDIC support.         D. Offiler
!   1.02     Jan 2004  Increase max. message length from 15,000
!                      to 25,000 bytes.                  D. Offiler
!   1.03     Mar 2005  Support MetDB v16.1 DEBUFR option to return
!                      Section 2 data (unused).          D. Offiler
!   1.05     May 2008  Make 'Message' string length definition
!                      use 'nb' parameter value from module
!                      instead of hard-coded value here. D. Offiler
!   1.06     Aug 2008  MetDB DEBUFR interface no longer supports
!                      Section 2 data - remove associated variable
!                      definitions & arguments from call. D. Offiler      
!

! Modules

  USE DecBUFRmod

  INCLUDE 'portability.fi'
  
! Fixed values
  
  CHARACTER (LEN=*), PARAMETER :: Fmt1 = &
                     "(1X,'BUFR Message',I4,':',I5,' observation(s)')"
  CHARACTER (LEN=*), PARAMETER :: Fmt2 = &
                     "(1X,I6,' BUFR Message(s) ',A)"
! Local variables

  CHARACTER (LEN=nb)    :: Message           ! BUFR message string
  CHARACTER (LEN=32767) :: Names             ! Decoded list of character data
  CHARACTER (LEN=100)   :: BFRdsn            ! Name of input file
  CHARACTER (LEN=100)   :: Outdsn            ! Name of output file
  REAL                  :: Values(nv)        ! Decoded values array
  INTEGER               :: Descr(nd)         ! Expanded descriptor list
  INTEGER               :: ierr              ! Error code
  INTEGER               :: Nbytes            ! No. of bytes decoded
  INTEGER               :: Ndesc             ! Local no. of descriptors
  INTEGER               :: Nobs              ! Local no. of observations
  INTEGER               :: i                 ! Loop counter
  INTEGER               :: Fmsgr, Nmsgr      ! First & no. of messages to read
  INTEGER               :: Nmsgrd = 0        ! No. of BUFR Messages read
  INTEGER               :: Nmsgde = 0        ! No. of BUFR Messages decoded
  INTEGER               :: Nmsgsk = 0        ! No. of BUFR Messages skipped
  INTEGER               :: ExitErr           ! Exit error flag
  LOGICAL               :: disphdrs          ! Display headers flag
  LOGICAL               :: dispsec4          ! Display Section 4 flag

! Get command line arguments & options

  CALL DecBUFRoptions ( BFRdsn,   &
                        Outdsn,   &
                        disphdrs, &
                        dispsec4, & 
                        Fmsgr,    &
                        Nmsgr,    &
                        ierr )
  IF ( ierr /= 0 ) THEN
    ExitErr = ErrIO
    CALL EXIT (ExitErr)
  ELSE
    ExitErr = ErrOK
  END IF
  
! Attempt to open output file if wanted.
! (Input file is opened by first call to BUFRREAD)

  IF ( Outdsn /= " " ) THEN
    OPEN ( UNIT=Ounit,     &
           FILE=Outdsn,    &
         STATUS="UNKNOWN", &
         IOSTAT=ierr )
    IF ( ierr /= 0 ) Outdsn = " "
  END IF

! Skip any initial unwanted messages in input file

  IF ( Fmsgr > 1 ) THEN
    ierr = 0
    WRITE ( *, * ) "Skipping", Fmsgr-1, " message(s)..."
    DO WHILE ( ierr == 0 .AND. &
              Nmsgsk < Fmsgr-1 )

      CALL BUFRREAD ( TRIM(BFRdsn), &
                      Message,      &
                      Nbytes,       &
                      ierr )

      IF ( ierr > 0 ) THEN
        WRITE ( *, * ) "*** Error reading BUFR input file "//TRIM(BFRdsn)
        SELECT CASE (ierr)
          CASE (1)
            WRITE ( *, * ) "    - Failed to open file"
          CASE (2)
            WRITE ( *, * ) "    - No data read from file"
          CASE (3)
            WRITE ( *, * ) "    - End-of-message tag '7777' not found"// &
                           " for message", Nmsgsk+1
          CASE DEFAULT
            WRITE ( *, * ) "    - Unknown error code",ierr
        END SELECT 
        ExitErr = ErrIO
        EXIT
      ELSE IF ( ierr < 0 ) THEN
        WRITE ( *, * ) "*** End-of-file"
        EXIT
      END IF

      Nmsgsk = Nmsgsk + 1
    END DO                     ! end skipping

  END IF                       ! skip check

! Loop over required messages in input file

  DO WHILE ( ierr == 0 .AND. &
             Nmsgde < Nmsgr )

    CALL BUFRREAD ( TRIM(BFRdsn), &
                    Message,      &
                    Nbytes,       &
                    ierr )

    IF ( ierr > 0 ) THEN
      WRITE ( *, * ) "*** Error reading BUFR input file "//TRIM(BFRdsn)
      SELECT CASE (ierr)
        CASE (1)
          WRITE ( *, * ) "    - Failed to open file"
        CASE (2)
          WRITE ( *, * ) "    - No data read from file"
        CASE (3)
          WRITE ( *, * ) "    - End-of-message tag '7777' not found"// &
                         " for message", Nmsgsk+Nmsgrd+1
        CASE DEFAULT
          WRITE ( *, * ) "    - Unknown error code",ierr
      END SELECT
      ExitErr = ErrIO
      EXIT

    ELSE IF ( ierr < 0 ) THEN
      WRITE ( *, "(/A)" ) " *** Normal end-of-file"
      EXIT
            
    ELSE IF ( INDEX ( Message, "BUFR" ) == 0 ) THEN
      WRITE ( *, * ) "*** Start-of-message tag 'BUFR' not found"// &
                     " for message", Nmsgsk+Nmsgrd+1
      WRITE ( *, * ) "    [", Nbytes, (ichar(Message(i:i)),i=1,20),"]"
    END IF

    Nmsgrd = Nmsgrd + 1
      
! Decode Message, with optional dump

    IF ( disphdrs ) CALL AnalBUFR ( Message )
    
    IF ( dispsec4 ) WRITE ( *, "(/A/)" ) " Data in Section 4:"
    Ndesc = nd
    Nobs  = nv
    CALL DEBUFR ( Descr,    &
                  Values,   &
                  Names,    &
                  Ndesc,    &
                  Nobs,     &
                  Message,  &
                  dispsec4 )

! Optionally output decoded values

    Nmsgde = Nmsgde + 1
    IF ( Outdsn /= " " ) THEN
      WRITE ( Ounit, "(/A/)" ) " Data in Section 4:"
      WRITE ( Ounit, FMT=Fmt1 ) Nmsgsk+Nmsgde, Nobs

      CALL OutputVals ( Values, &
                        Descr,  &
                        Names,  &
                        Ndesc,  &
                        Nobs )
    END IF
    
  END DO                    ! processing messages

  IF ( Outdsn /= " " ) THEN
    CLOSE ( UNIT=Ounit, &
          IOSTAT=ierr )
  END IF
  
  WRITE ( *, FMT=Fmt2 ) Fmsgr-1, "skipped"
  WRITE ( *, FMT=Fmt2 ) Nmsgrd,   "read"
  WRITE ( *, FMT=Fmt2 ) Nmsgde,   "decoded"
     
! Signal final exit error code

  CALL EXIT(ExitErr)

END PROGRAM DecBUFR
!-------------------------------------------------------------------------
SUBROUTINE DecBUFRoptions ( BFRdsn,   & ! (out)
                            Outdsn,   & ! (out)
                            disphdrs, & ! (out)
                            dispsec4, & ! (out)
                            Fmsgr,    & ! (out)
                            Nmsgr,    & ! (out)
                            ierr )     ! (out)
!
! Purpose:
!   Command line options interface for generic BUFR decoder 
!
! Arguments:
!   BFRdsn   chr/out  BUFR input file name (required)
!   Outdsn   chr/out  data output file name (optional)
!   disphdrs log/out  decoded headers dump   display (default: no)
!   dispsec4 log/out  decoded Section 4 dump display (default: no)
!   Fmsgr    int/out  first message to read (default: 1)
!   Nmsgr    int/out  no. of messages to read (default: all)
!   ierr     int/out  exit status:
!                     0 = ok
!                     1 = help requested
!                     2 = input file not specified
!                     3 = input file not found
!
! Calls:
!   GETARG
!   IARGC
!
! Called by:
!   DecBUFR
!
! Modules:
!   DecBUFRmod  - fixed parameter definitions
!
! Includes:
!   portability.fi  - system dependent settings
!
! Description:
!   Command line interface:
!     > decbufr bufr_file [-d[h]]   [-o output_file]
!                         [-f first | -s skip] 
!                         [-l last  | -n number]
!                         [-h | ?]
!    where:
!      bufr_file   is the input file containing BUFR message(s)
!      -d          writes an interpreted diagnostic dump to stdout
!      -dh         writes an interpreted headers dump only to stdout
!      output_file is a (formatted text) data file
!      first       is the first message to decode (ie skip first-1)
!      skip        is the number of messages to skip (ie first=skip+1)
!      last        is the last message to decode (ie number=first-last+1)
!      number      is the no. of messages to decode from first.
!    Arguments can be in any order and any space(s) between a
!    switch and its value is optional. If both -f & -s or both 
!    -l & -n options are present, only the second is taken. 
!    If the bufr_file is not specified, or if '?' or '-h', options
!    are present, a usage reminder only is printed and no decoding
!    is done.
!
! History:
!   Version  Date      Change                           By
!  -----------------------------------------------------------------
!   1.00     Dec 2001  First f90 release                D. Offiler
!   1.01     Jun 2002  Add version number to help o/p   D. Offiler
!   1.02     Aug 2003  Use portability.fi               D. Offiler
!
! Modules

  USE DecBUFRmod, ONLY: Version

! Includes

  INCLUDE 'portability.fi'

! Argument list parameters 

  CHARACTER (LEN=*), INTENT(OUT) :: BFRdsn   ! BUFR (input)    dataset name
  CHARACTER (LEN=*), INTENT(OUT) :: Outdsn   ! Decode (output) dataset name
  INTEGER,           INTENT(OUT) :: Fmsgr    ! First message to decode
  INTEGER,           INTENT(OUT) :: Nmsgr    ! Number of messages to decode
  INTEGER,           INTENT(OUT) :: ierr     ! Return status value
  LOGICAL,           INTENT(OUT) :: disphdrs ! Display headers on/off flag
  LOGICAL,           INTENT(OUT) :: dispsec4 ! Display Section 4 on/off flag

! Local parameters

  CHARACTER (LEN=100) :: arg     ! command line argument
  CHARACTER (LEN=100) :: Value   ! value extracted from arg
  INTEGER             :: i, n    ! counters
  INTEGER             :: Narg    ! number of command line arguments
  LOGICAL             :: exists  ! file exists flag

! Set up

  BFRdsn   = " "
  Outdsn   = " "
  disphdrs = .FALSE.
  dispsec4 = .FALSE.
  Fmsgr    = 1
  Nmsgr    = 999999
  ierr     = 0

! Process command line arguments & options. Defaults are no dump or decoded
! output, processing all messages. There is no default for input file name.

  Narg = IARGC()
  i = 1

  DO WHILE ( i <= Narg )
    CALL GETARG ( i, arg )

! Help wanted

    IF ( arg(1:1) == "?"  .OR. &
         arg(1:2) == "-h" .OR. &
         arg(1:2) == "-H" ) THEN
      ierr = 1

! Dump output (-d for all sections, -dh for headers only)

    ELSE IF ( arg(1:2) == "-d"  .OR. &
              arg(1:2) == "-D" ) THEN
      disphdrs = .TRUE.
      IF ( arg(3:3) /= "h" .AND. &
           arg(3:3) /= "H" ) dispsec4 = .TRUE.
      
! Output file

    ELSE IF ( arg(1:2) == "-o"  .OR. &
              arg(1:2) == "-O" ) THEN
      Outdsn = arg(3:)
      IF ( Outdsn == " " ) THEN
        i = i + 1
        CALL GETARG ( i, Outdsn )
      END IF

! Start Message (first)

    ELSE IF ( arg(1:2) == "-f"  .OR. &
              arg(1:2) == "-F" ) THEN
      Value = arg(3:)
      IF ( Value == " " ) THEN
        i = i + 1
        CALL GETARG ( i, Value )
      END IF
      READ ( Value, *, IOSTAT=ierr ) n
      IF ( ierr == 0 .AND. &
           n    >  1 ) Fmsgr = n

! Start Message (skip)

    ELSE IF ( arg(1:2) == "-s"  .OR. &
              arg(1:2) == "-S" ) THEN
      Value = arg(3:)
      IF ( Value == " " ) THEN
        i = i + 1
        CALL GETARG ( i, Value )
      END IF
      READ ( Value, *, IOSTAT=ierr ) n
      IF ( ierr == 0 .AND. &
           n    >  0 ) Fmsgr = n + 1

! Number of Messages

    ELSE IF ( arg(1:2) == "-n"  .OR. &
              arg(1:2) == "-N" ) THEN
      Value = arg(3:)
      IF ( Value == " " ) THEN
        i = i + 1
        CALL GETARG ( I, Value )
      END IF
      READ ( Value, *, IOSTAT=ierr ) n
      IF ( ierr == 0 .AND. &
           n    >  0 ) Nmsgr = n

! Last Message

    ELSE IF ( arg(1:2) == "-l"  .OR. &
              arg(1:2) == "-L" ) THEN
      Value = arg(3:)
      IF ( Value == " " ) THEN
        i = i + 1
        CALL GETARG ( i, Value )
      END IF
      READ ( Value, *, IOSTAT=ierr ) n
     IF ( ierr == 0 .AND. &
          n    >  0 ) Nmsgr = n + Fmsgr - 1

! If not a (unknown) switch, assume it's the input file Name

    ELSE IF ( arg(1:1) /= "-" ) THEN
      BFRdsn = arg

! Ignore anything else

    ELSE
      CONTINUE
      
    END IF

    i = i + 1    ! next arguent
    
  END DO         ! argument loop

! Processed all arguments - was input file name given?

  IF ( BFRdsn == " " ) THEN 
    ierr = 2
  ELSE
    INQUIRE ( FILE=BFRdsn, EXIST=exists )
    IF ( .NOT. exists ) THEN
      ierr = 3
      WRITE ( *, * ) "*** Input BUFR file ", TRIM(BFRdsn), " not found"
    END IF
  END IF
  
! If input file not specified or help wanted, give a usage reminder

  IF ( ierr >= 1 .AND. &
       ierr <= 2 ) THEN 
    WRITE ( *, * ) "Usage: decbufr bufr_file"
    WRITE ( *, * ) "               [-d[h]] [-o decode_file]"
    WRITE ( *, * ) "               [-f first | -s skip]"
    WRITE ( *, * ) "               [-l last  | -n number]"
    WRITE ( *, * ) " Defaults: bufr_file   - required"
    WRITE ( *, * ) "           dump output - no output"
    WRITE ( *, * ) "           decode_file - no output"
    WRITE ( *, * ) "           first       - 1  or"
    WRITE ( *, * ) "           skip        - 0"
    WRITE ( *, * ) "           last        - 999999  or"
    WRITE ( *, * ) "           number      - 999999"
    WRITE ( *, * ) " Version: " // Version
  END IF

END SUBROUTINE DecBUFRoptions
!-----------------------------------------------------------------
SUBROUTINE AnalBUFR ( Message ) ! (in)
!
! Purpose:
!   Analyse a BUFR message (except data in secton 4)
!
! Arguments:
!   Message  chr/in   BUFR bit string
!
! Calls:
!   TableA
!   CT001033
!   CT001034
!   DESFXY
!   VALUE
!
! Called by:
!   DecBUFR
!
! Description:
!   Analyses a BUFR (edition 1,2,3) message; checks all sections
!   but skips any data in Section 2 (just flags whether present or
!   not). The actual data in section 4 is not interpreted. Info is
!   written to stdout.
!
! References:
!   Manual on codes. WMO-306, Vol 1.2, Part B, 1995 with suppliment 
!   no.3, Aug 1998 and updates Nov 2000.
!
! History:
!   Version  Date      Change                       By
!  -----------------------------------------------------------------
!   1.00     Dec 2001  First f90 release            D. Offiler
!   1.01     Mar 2003  Substitute 001031 for 001033 D. Offiler
!   1.02     Apr 2004  Fall back on separate local table for
!                      001034 (Orig. Sub-Centre) if not defined
!                      in main Code/Flag Tables.    D. Offiler
!   1.03     Jul 2006  Use separate local tables for
!                      001031/001033 (Orig/Gen Centre) and
!                      001034 (Orig. Sub-centre)    D. Offiler
!
  USE codetables
  
  IMPLICIT NONE

! Fixed values

  CHARACTER (LEN=3), PARAMETER :: Months(0:12) = &
                     (/ "???","Jan","Feb","Mar","Apr","May","Jun", &
                              "Jul","Aug","Sep","Oct","Nov","Dec" /)
                              
  CHARACTER (LEN=*), PARAMETER :: Fmt1 = &
                     "(/1X,'BUFR Section',I2 )"
  CHARACTER (LEN=*), PARAMETER :: Fmt2 = &
                     "( 2X,'Edition:',I2/" // &
                       "2X,'Total BUFR Message length: ',A,' octets' )"
  CHARACTER (LEN=*), PARAMETER :: Fmt3 = &
                     "( 2X,'BUFR Master Table:     ',I4 )"
  CHARACTER (LEN=*), PARAMETER :: Fmt4 = &
                     "( 2X,'Originating Centre:    ',I4,' (',A,')'/" // &
                       "2X,'Originating Sub-Centre:',I4,' (',A,')'/" // &
                       "2X,'Update Sequence No:    ',I4 )"
  CHARACTER (LEN=*), PARAMETER :: Fmt5 = &
                     "( 2X,'Data Category:         ',I4,' (',A,')'/" // &
                       "2X,'Data sub-Category:     ',I4/" // &
                       "2X,'Table version:       ',I6 )"
  CHARACTER (LEN=*), PARAMETER :: Fmt6 = &
                     "( 2X,'Data Category:         ',I4,' (',A,')'/" // &
                       "2X,'Data sub-Category:     ',I4/" // &
                       "2X,'Master Table:          ',I4,'  Local Table:  ',I4 )"
  CHARACTER (LEN=*), PARAMETER :: Fmt7 = &
                     "( 2X,'Message time:         ',I2.2,':',I2.2,1X," // &
                                                   "I2.2,'-',A3,'-',I4 )"
  CHARACTER (LEN=*), PARAMETER :: Fmt8 = &
                     "( 2X,'No. of observations expected in Section 4: ',A )"
  CHARACTER (LEN=*), PARAMETER :: Fmt9 = &
                     "( 2X,'Descriptor', I4,': ',I1,I2.2,I3.3 )"
                     
! Argument list parameters

  CHARACTER (LEN=*), INTENT(IN)  :: Message   ! BUFR message bit string
  
! Local parameters

  CHARACTER (LEN=30) :: Category  ! Table A data category
  CHARACTER (LEN=30) :: Centre    ! Centre name
  CHARACTER (LEN=30) :: SubCentre ! Sub-centre name
  CHARACTER (LEN=4)  :: ICAO      ! Centre ICAO code
  CHARACTER (LEN=4)  :: Delim     ! Delimiter
  CHARACTER (LEN=8)  :: number    ! Number representation
  
  INTEGER :: Ndesc                ! No. of descriptors in Section 3
  INTEGER :: Nobs                 ! No. of observations in Section 3
  INTEGER :: Descr                ! Descriptor(s) in Section 3
  INTEGER :: DateTime(7)          ! Date & time in Section 1
  INTEGER :: DataType(2)          ! Data type in Section 1
  INTEGER :: F, X, Y              ! Descriptor parts
  INTEGER :: i                    ! Loop counter
  INTEGER :: ib, id, idata        ! Tempory values
  INTEGER :: is2                  ! Is Section 2 present? flag
  INTEGER :: Edition              ! BUFR Edition in Section 0
  INTEGER :: BufrMasterTable      ! Master table ID in Section 1
  INTEGER :: OrigCentre           ! Orig. centre code in Section 1
  INTEGER :: OrigSubCentre        ! Orig. sub-centre code in Section 1
  INTEGER :: UpdtSeqNo            ! Update sequence no. in Section 1
  INTEGER :: VerMasterTable       ! Master table version in Section 1
  INTEGER :: VerLocalTable        ! Local table version in Section 1
  INTEGER :: L                    ! Pointer to current byte pos. in Message
  INTEGER :: Byte1                ! Pointer to 'BUFR' in Message
  INTEGER :: len0, len1, len2     ! Lengths (bytes) of Sections 0, 1, 2
  INTEGER :: len3, len4, lenm     ! Lengths (bytes) of Sections 3, 4 & total
  INTEGER :: nm = 0               ! Sequence no. of this message

! Functions

  INTEGER :: IDES                 ! function to convert F,X,Y-->descr
  INTEGER :: VALUE                ! function to extract bits value

  SAVE nm
  
! set up

  nm = nm + 1
  WRITE ( number, "(I8)" ) nm 
  WRITE ( *, "(/2A)" ) "   BUFR Decoder Diagnostic Dump - Message no. ", &
                       ADJUSTL ( number )
  lenm = LEN ( Message )

! ---- Section 0 ('BUFR') ----

  WRITE ( *, FMT=Fmt1 ) 0

! Check for start of message delimiter

  Delim = "BUFR"
  L = INDEX ( Message, Delim )
  IF ( L == 0 ) THEN             
    WRITE ( *, * ) "*** Message Delimiter 'BUFR' not present"
    RETURN
  END IF
  Byte1 = L

! Look at octet 8; if the value is 0 or 1, then this is
! an Edition 1 Message (Section 1 starts in octet 5) else Edition 2 or 3
! (Section 1 starts in octet 9) and total Message length is in octets 5-7.
! For Edition 1, we assume Message length is length of Message string.
 
  Edition = ICHAR ( Message(L+7:L+7) )
  IF ( Edition <= 1 ) THEN
    Edition = 1
    len0    = 4
  ELSE
    len0    = 8
    ib      = ( Byte1 + 3 ) * 8 
    lenm    = VALUE ( Message, ib, 24 )
  END IF
  WRITE ( number, "(I8)" ) lenm
  WRITE ( *, FMT=Fmt2 ) Edition, &
                        TRIM(ADJUSTL(number))
  L = L + len0 - 1

! ---- Section 1 (identification) ----

  WRITE ( *, FMT=Fmt1 ) 1

  ib = L * 8
  len1 = VALUE ( Message, ib, 24 )
  IF ( len1+L-Byte1 > lenm ) THEN
    WRITE ( *, * ) "*** Section length greater than Message length"
    RETURN
  END IF

  IF ( Edition >= 2 ) THEN
    BufrMasterTable = ICHAR ( Message(L+4:L+4) )
    WRITE ( *, FMT=Fmt3 ) BufrMasterTable
  END IF

! Orginating Centre & Sub-Centre and Update Sequence Number

  OrigSubCentre = ICHAR ( Message(L+5:L+5) )
  OrigCentre    = ICHAR ( Message(L+6:L+6) )
  UpdtSeqNo     = ICHAR ( Message(L+7:L+7) )
  
  CALL CT001033 ( OrigCentre, Centre, ICAO ) 
  IF ( OrigSubCentre > 0 ) THEN
    CALL CT001034 ( OrigSubCentre, SubCentre )
  ELSE
    SubCentre = "None"
  END IF
  WRITE ( *, FMT=Fmt4 ) OrigCentre,    TRIM(ADJUSTL(Centre)),    &
                        OrigSubCentre, TRIM(ADJUSTL(SubCentre)), &
                        UpdtSeqNo
! Data Type

  is2            = ICHAR ( Message(L+8:L+8) ) 
  DataType(1)    = ICHAR ( Message(L+9:L+9) )
  DataType(2)    = ICHAR ( Message(L+10:L+10) )
  VerMasterTable = ICHAR ( Message(L+11:L+11) )
  VerLocalTable  = ICHAR ( Message(L+12:L+12) )
  
  CALL TableA ( DataType(1), Category )
  IF ( Edition <= 1 ) THEN
    WRITE ( *, FMT=Fmt5 ) DataType(1), TRIM(Category), &
                          DataType(2), &
                          VerMasterTable*256+VerLocalTable
  ELSE   
    WRITE ( *, FMT=Fmt6 ) DataType(1), TRIM(Category), &
                          DataType(2), &
                          VerMasterTable, VerLocalTable
  END IF

! Date & Time (pivot 2-digit century year to CCYY)

  DO i = 1, 5
    DateTime(i) = ICHAR ( Message(L+I+12:L+I+12) )
  END DO
  IF ( DateTime(1) < 50 ) THEN
    DateTime(1) = DateTime(1) + 2000
  ELSE IF ( DateTime(1) < 100 ) THEN
    DateTime(1) = DateTime(1) + 1900
  ELSE IF ( DateTime(1) == 100 ) THEN
    DateTime(1) = 2000
  END IF
  i = DateTime(2)
  IF ( i < 0 .OR. i > 12 ) i = 0
  WRITE ( *, FMT=Fmt7 ) DateTime(4), DateTime(5), &
                        DateTime(3), Months(i), DateTime(1)
  L = L + len1

! ---- Optional Section 2 (local data) ----

  WRITE ( *, FMT=Fmt1 ) 2

  IF ( BTEST ( is2, 7 ) ) THEN
    ib     = L * 8
    len2   = VALUE ( Message, ib, 24 )
    IF ( len2+L-Byte1 > lenm ) THEN
      WRITE ( *, * ) "*** Section length greater than Message length" 
      RETURN
    END IF
    WRITE ( *, "(A,I6,A)" ) " Section 2 present but skipped - ", &
                            len2, " octets"
  ELSE                 
    len2   = 0
    WRITE ( *, * ) " Section 2 not present"
  END IF
  L = L + len2

! ---- Section 3 (Descriptors) ----

  WRITE ( *, FMT=Fmt1 ) 3

  ib = L * 8
  len3 = VALUE ( Message, ib, 24 )
  IF ( len3+L-Byte1 > lenm ) THEN
    WRITE ( *, * ) "*** Section length greater than Message length" 
    RETURN
  END IF

! Number of obs. in Section 4

  ib = ( L + 4 ) * 8
  Nobs = VALUE ( Message, ib, 16 )
  IF ( Nobs == 0 ) THEN
    WRITE ( *, * ) "*** No observations present"
    RETURN
  END IF

! Observed data or not?

  idata = ICHAR ( Message(L+7:L+7) )
  IF ( BTEST ( idata, 7 ) ) THEN
    WRITE ( *, * ) " Message contains observed data"
  ELSE
    WRITE ( *, * ) " Message contains non-observed data"
  END IF 
  WRITE ( number, "(I8)" ) Nobs
  WRITE ( *, FMT=Fmt8 ) ADJUSTL(number)

! Compressed or not?

  IF ( BTEST ( idata, 6 ) ) THEN
    WRITE ( *, * ) " Data is compressed"
  ELSE
    WRITE ( *, * ) " Data is not compressed"
  END IF

! No. of descriptors & list

  Ndesc = ( len3 - 7 ) / 2
  IF ( Ndesc == 0 ) THEN
    WRITE ( *, * ) "*** No Descriptors present"
    RETURN  
  END IF

  ib = ( L + 7 ) * 8
  DO id = 1, Ndesc
    Descr = VALUE ( Message, ib, 16 )
    CALL DESFXY ( Descr, F,X,Y )
    WRITE ( *, FMT=Fmt9 ) id, F,X,Y
  END DO

  L = L + len3        

! ---- Section 4 (data) ----

  WRITE ( *, FMT=Fmt1 ) 4

  ib = L * 8
  len4 = VALUE ( Message, ib, 24 )
  IF ( len4+L-Byte1 > lenm ) THEN
    WRITE ( *, * ) "*** Section length greater than Message length" 
    RETURN
  ELSE
    WRITE ( number, "(I8)" ) len4
    WRITE ( *, "(3A)" ) "  Data length: ", TRIM(ADJUSTL(number)), " octets"
  END IF

! ---- Section 5 ('7777') ----

  WRITE ( *, FMT=Fmt1 ) 5
  
! Check for end-of-message delimiter

  Delim = "7777"
  IF ( Edition >= 3 ) THEN
    L = lenm - 4
    IF ( Message(L+1:L+4) /= Delim ) THEN
      WRITE ( *, * ) "*** Message Delimiter '7777' not present"
    ELSE
      WRITE ( *, * ) " OK"
    END IF
  ELSE
    L = INDEX ( Message, Delim )
    IF ( L == 0 ) THEN
      WRITE ( *, * ) "*** Message Delimiter '7777' not present"
    ELSE
      WRITE ( *, * ) " OK"
    END IF
  END IF

END SUBROUTINE AnalBUFR         
!-----------------------------------------------------------------
SUBROUTINE OutputVals ( Values, & ! (in)
                        Descr,  & ! (in)
                        Names,  & ! (in)
                        Ndesc,  & ! (in)
                        Nobs )    ! (in)
!
! Purpose:
!   Output BUFR Section 4 decoded data for one message.
!
! Arguments:
!   Values  int  Array of decoded values
!   Descr   int  Array of expanded descriptors
!   Names   chr  String of character data
!   Ndesc   int  No. of descriptors
!   Nobs    int  No. of observations
!
! Calls:
!   DESFXY
!   TABLEB
!
! Called by:
!   DecBUFR
!
! Modules:
!   DecBUFRmod  - fixed parameter definitions
!
! Called by:
!   DecBUFR
!
! Description:
!   Outputs decoded BUFR message values in a simple list format.
!   Output is written to stream #8
!
! References:
!   Manual on codes. WMO-306, Vol 1.2, Part B, 1995 with suppliment 
!   no.3, Aug 1998 and updates Nov 2000.
!
! History:
!   Version  Date      Change                                 By
!  -----------------------------------------------------------------
!   1.00     Dec 2001  First f90 release                      D. Offiler
!   1.01     Jun 2002  No changes to this routine
!   1.02     Jun 2002  Signal exit error code.                D. Offiler
!   1.03     May 2004  Subroutine TABLEB in MetDB kernel package v15
!                      now returns FORMAT as 'F' for flag table and
!                      'C' for code table (used to be 'F' for both).
!                      Tests modified to use this information. D. Offiler
!   1.04     Aug 2006  For scale=0, output was in integer format (I11);
!                      Some very large numbers (of order 10^14) are 
!                      out of range of I*4 so not correctly printed.
!                      Now modified so that any number >1E10 is
!                      output as floating-point E-format.      D. Offiler
!

! Modules

  USE DecBUFRmod
  
  INCLUDE 'portability.fi'

! Fixed values

  CHARACTER (LEN=*), PARAMETER :: Fmt1 = "(1X,I6,I2,I2.2,I3.3,I11)"
  CHARACTER (LEN=*), PARAMETER :: Fmt2 = "(1X,I6,I2,I2.2,I3.3,1X,A)"
  CHARACTER (LEN=*), PARAMETER :: Fmt4 = "(1X,I6,I2,I2.2,I3.3,E15.7)"
  
! Argument list parameters

  CHARACTER (LEN=*), INTENT(IN) :: Names       ! Character data
  REAL,              INTENT(IN) :: Values(nv)  ! Decoded values
  INTEGER,           INTENT(IN) :: Descr(nd)   ! Expanded list of descriptors
  INTEGER,           INTENT(IN) :: Ndesc       ! No. of descriptors
  INTEGER,           INTENT(IN) :: Nobs        ! No. of observations

! Local variables

  CHARACTER (LEN=1)  :: Format                 ! Table B element format type
  CHARACTER (LEN=60) :: Name                   ! Table B element name
  CHARACTER (LEN=24) :: Units                  ! Table B element units
  CHARACTER (LEN=30) :: Fmt3                   ! Format descriptor
  INTEGER            :: i, j                   ! Loop counters
  INTEGER            :: F,X,Y                  ! Descriptor parts
  INTEGER            :: Dcount                 ! Count of element descriptors
  INTEGER            :: Ocount                 ! Count of observations
  INTEGER            :: Scale                  ! Table B scale
  INTEGER            :: RefVal                 ! Table B reference value
  INTEGER            :: Width                  ! Table B bit width
  INTEGER            :: AddScale = 0           ! Temporary Scale change
  INTEGER            :: pt                     ! Pointer to start of chr data
  INTEGER            :: nc                     ! No. of chrs in chr data
  DOUBLE PRECISION   :: Rval                   ! Temporary value
  INTEGER            :: Ival                   ! Temporary value

  DOUBLE PRECISION, ALLOCATABLE :: UserVals(:,:) ! Temporary table
  INTEGER                       :: AllocStat     ! Allocation status
  
! Allocate temporary space for re-arranged obs. store

  IF ( .NOT. ALLOCATED(UserVals) ) THEN
    ALLOCATE(UserVals(1:Nobs,1:Ndesc), STAT=AllocStat)
    IF ( AllocStat /= 0 ) THEN
      WRITE ( *, * ) "*** Failed to allocated UserVals array"
      CALL EXIT(ErrMem)
    END IF
  END IF
  
! Copy data array to (obs,elem) order

  Ocount = 0
  DO j = 1, Ndesc
    DO i = 1, Nobs
      Ocount = Ocount + 1
      UserVals(i,j) = Values(Ocount)
    END DO
  END DO

! Loop over all observations...

  DO i = 1, Nobs
    WRITE ( Ounit, "(/1X,'Observation number: ',I5/)") i
    Dcount = 0

! Loop over descriptors for this ob....

    DO j = 1, Ndesc
      CALL DESFXY ( Descr(j), F,X,Y )

      IF ( F == 0 ) THEN                   ! element descriptor
        Dcount = Dcount + 1
        Rval = UserVals(i,Dcount)
        Ival = NINT ( Rval )
            
        IF ( Descr(j) < 65536 ) THEN       ! numeric element
          IF ( Ival /= Missing ) THEN
            CALL TABLEB ( X, Y, Scale,  RefVal, Width, &
                                Format, Name,   Units )
            IF ( Format == "C" .OR.  &
                 Format == "F" ) THEN
              WRITE ( Ounit, FMT=Fmt1 ) j, F,X,Y, Ival
            ELSE
              Scale = MAX ( Scale, AddScale )
              IF ( Scale > 0 ) THEN
                 IF ( Rval < 1.0E10 ) THEN
                   WRITE ( Fmt3, "(A,I2,A,I1,A)" ) &
                          "(1X,I6,I2,I2.2,I3.3,F", 12+Scale, ".", Scale, ")"
                   WRITE ( Ounit, FMT=Fmt3 ) j, F,X,Y, Rval
                 ELSE
                   WRITE ( Ounit, FMT=Fmt4 ) j, F,X,Y, Rval
                 END IF
              ELSE
                 IF ( Rval < 1.0E10 ) THEN
                   WRITE ( Ounit, FMT=Fmt1 ) j, F,X,Y, Ival
                 ELSE
                   WRITE ( Ounit, FMT=Fmt4 ) j, F,X,Y, Rval
                 END IF          
              END IF
            END IF
          ELSE
            WRITE ( Ounit, FMT=Fmt1 ) j, F,X,Y, Missing
          END IF
              
        ELSE                                ! pointer to chr element
          pt = MOD ( Ival, 65536 )
          nc = Ival / 65536
          WRITE ( Ounit, FMT=fmt2 ) j, F,X,Y, Names(pt:pt+nc-1)
        END IF
     
      ELSE                                  ! descr other than element
        WRITE ( Ounit, FMT=Fmt1 ) j, F,X,Y
        IF ( F == 2 .AND. X == 2 ) THEN
          IF ( Y > 0 ) THEN
            AddScale = Y - 128
          ELSE
            AddScale = 0
          END IF
        END IF
            
      END IF            ! descr type
                  
    END DO              ! descr loop
        
  END DO                ! nobs loop

! De-allocate temporary array

  IF ( ALLOCATED(UserVals) ) THEN
    DEALLOCATE(UserVals)
  END IF


END SUBROUTINE OutputVals
