!------------------------------------------------------------------------------
!
!   GTSHDRS.F90   --   WMO/GTS HEADER/TRAILER ROUTINES FOR BUFR MESSAGES
!
!------------------------------------------------------------------------------
!
!   CALL GTSHDR   ( GTShdrType, ITseq, TTA, Lat, Lon, II, CCCC, DateTime, 
!                   Message, LenHdr )
!
!   CALL GTSEND   ( GTShdrType, Message, LenMsg )
!
!------------------------------------------------------------------------------
! (c) CROWN COPYRIGHT 2007, METEOROLOGICAL OFFICE, All Rights Reserved.
!
! Use, duplication or disclosure of this code is subject to the
! restrictions as set forth in the contract.
!
!                The Met. Office, London Road,
!                BRACKNELL, Berkshire, UK, RG12 2SZ
!
! If no contract has been raised with this copy of the code, the use,
! duplication and disclosure of it is strictly prohibited. Permission
! to do so must first be obtained in writing, from the Head of Satellite
! Applications at the above address.
!
!--------------------------------------------------------------------
SUBROUTINE GTShdr ( GTShdrType, & ! (in)
                    ITseq,      & ! (in)
                    TTA,        & ! (in)
                    Lat,        & ! (in)
                    Lon,        & ! (in)
                    II,         & ! (in)
                    CCCC,       & ! (in)
                    DateTime,   & ! (in)
                    Message,    & ! (out)
                    LenHdr )      ! (out)
!
! Purpose:
!   Generate WMO abbreviated GTS header sequence
!
! Arguments:
!  GTShdrType  int  Code for leading GTS header type:
!                    <= 0 no header at all
!                     = 1 header only, no leading bytes
!                     = 2 include 4 leading null bytes  (X.25)
!                     = 3 include 10 '0's for size/type (TCP/IP)
!  ITseq       int  Message transmission sequence number (1-999)
!  TTA         chr  3-chr data ID 1 (t1t2a1)
!  Lat         int  Latitude  of data (-90  to +90 deg.)
!  Lon         int  Longitude of data (-180 to +180 or 0 to 360 deg.)
!  II          chr  2-chr data ID 2 (ii)
!  CCCC        chr  4-chr GTS source node ID (cccc)
!  DateTime    int  5-element date/time (yr,mon,day,hr,min)
!  Message     chr  Coded header string
!  LenHdr      int  Total length of header   
!
! Description:
!   Codes data routing header information into WMO/GTS standard 
!   31-byte abbreviated heading of the form: 
!     'sssrrriii'
!   where:
!     'sss' is a 10-byte message start sequence: 'SOH/CR/CR/LF/NNN/CR/CR/LF' 
!     'rrr' is the 18-byte routing header:       't1t2a1a2ii cccc YYGGgg'
!     'iii' is a 3-byte data introducer iii:     'CR/CR/LF'.
!   The message sequence number (NNN) should be a sequential 3-digit 
!   integer cycling from 001 to 999, and is coded from ITseq.
!   Data ID (t1t2a1 and ii) is coded as given. Geographical area code 
!   (a2) is derived from given lat/lon of data. Source node (cccc) is 
!   the centre coding the data (e.g. 'EGRR' for Bracknell). Day of month
!   (YY), hour (GG) and minute (gg) are coded from DateTime.
!   Optionally, prepend 4 null bytes (if GTShdrTypes=2) for X.25
!   support, or 10-bytes for IP support (if GTShdrType=3). No headers
!   are created at all if GTShdrType is not 1, 2 or 3. 
!   On exit, LenHdr is the total number of bytes in the header
!   (min:0, max:41) - (i.e. the BUFR/GRIB message should start in byte 
!   LenHdr+1)
!   Finally, call GTSend to complete the GTS bulletin by appending a trailer
!   and filling the leader with the actual mesaage length.
!
! References:
!   1) Manual on the Global Telecommunications System.
!      Vol.1, Part II. WMO, Geneva, 1986 (and updates)
!   2) Operatonal Procedures for the GTS, Attachment II-5,
!      Data Designators T1T2A1A2ii in Abbreviated Headings,
!      WMO, Geneva, 2003.
!
! History:
!
! Version Date          Comment                          By
! ------- ----          -------                          ------------
! 1.0-0   10-Dec-2001   First release.                   D. Offiler
! 1.1-0   06-Jun-2002   Remove EBCDIC support            D. Offiler
! 1.1-1   13-Jun-2007   Allow given longitudes to be in 
!                       the range +/-180deg or 0-360deg. D. Offiler
!
! End of header ---------------------------------------------------------------
!

  IMPLICIT NONE

! Fixed values

  CHARACTER (LEN=*), PARAMETER :: ALPHAB    = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
  CHARACTER (LEN=1), PARAMETER :: zero      = "0"
  INTEGER,           PARAMETER :: start(10) = &
                                 (/ 1, 13, 13, 10, 30, 30, 30, 13, 13, 10 /)
  INTEGER,           PARAMETER :: intro(3)  = (/ 13, 13, 10 /)
  INTEGER,           PARAMETER :: nul       = 0
                
! Argument list parameters

  CHARACTER (LEN=3), INTENT(IN)  :: TTA
  CHARACTER (LEN=2), INTENT(IN)  :: II
  CHARACTER (LEN=4), INTENT(IN)  :: CCCC
  CHARACTER (LEN=*), INTENT(OUT) :: Message
  INTEGER,           INTENT(IN)  :: GTShdrType
  INTEGER,           INTENT(IN)  :: ITseq
  INTEGER,           INTENT(IN)  :: Lat, Lon
  INTEGER,           INTENT(IN)  :: DateTime(*)
  INTEGER,           INTENT(OUT) :: LenHdr

! Local parameters

  INTEGER :: i, j
  INTEGER :: LL

  Message = " "

  IF ( GTShdrType >= 1 .AND. &
       GTShdrType <= 3 ) THEN

! GTS Message start line and header introducer sequence, 
! with leading nulls (X.25) or room for size/type (IP) if required

    IF ( GTShdrType == 2 ) THEN
      DO i = 1, 4
        Message(i:i) = CHAR(nul)
      END DO
      j = 4

    ELSE IF ( GTShdrType == 3 ) THEN
      Message(1:10) = REPEAT(zero,10)
      j = 10

    ELSE
      j = 0

    END IF

    DO i = 1, 10
      Message(i+j:i+j) = CHAR ( start(i) )
    END DO
    j = j + 11
    WRITE ( Message(j-6:j-4), FMT="(I3.3)" ) MOD ( ITseq, 1000 )

! WMO/GTS abbreviated heading (T1T2A1)

    Message(j:j+2) = TTA

! Area code A-L (A2) for Longitude segments 0-90W, 90W-180, 180-90E, 90E-0
! and for Latitude bands 90N-30N, 30N-30S, 30S-90S, plus ID code and
! originator ICAO code.

    LL = 4 - MOD(Lon+360,360) / 90
    LL = MIN ( MAX ( LL, 1 ), 4 )
    IF ( Lat .LT. -30 ) THEN
       LL = LL + 8
    ELSE IF ( Lat .LT. 30 ) THEN
       LL = LL + 4                 
    END IF
    Message(j+3:j+11) = ALPHAB(LL:LL) // II // " " // CCCC // " "

! Day of month & time

    WRITE ( Message(j+12:j+17), FMT="(3I2.2)" ) DateTime(3:5)

! Data introducer sequence

    DO i = 1, 3
      Message(i+j+17:i+j+17) = CHAR ( intro(i) )
    END DO

    LenHdr = j + 20

! No headers wanted

  ELSE
    LenHdr = 0

  END IF
  
END SUBROUTINE GTShdr
!-----------------------------------------------------------------------
SUBROUTINE GTSend ( GTShdrType, & ! (in)
                    Message,    & ! (inout)
                    LenMsg )      ! (inout)
!
! Purpose:
!   Append WMO GTS end-of-message sequence
!
! Arguments:
!  GTShdrType  int  Code for leading GTS header type:
!                    <= 0 no header at all
!                     = 1 header only, no leading bytes
!                     = 2 include 4 leading null bytes
!                     = 3 include 10 '0's for size/type
!   Message    chr  message string
!   LenMsg     int  Length of message. On entry, first byte
!                   to start trailer; on exit, last byte of 
!                   message including trailer (if any).
!
! Description:
!   Adds WMO GTS 4-byte end-of-message sequence 'CR/CR/LF/ETX' 
!   to given string starting at given byte and sets any following bytes
!   to null. On exit, 'LenMsg' points to the last byte of the complete 
!   message (the ETX), but if there is insufficient room left in the 
!   string, LenMsg is returned as zero. A trailer is only appended if
!   GTShdrType is 1, 2 or 3.
!
! History:
!
! Version Date          Comment
! ------- ----          -------
! 1.0-0   10-Dec-2001   First release.            D. Offiler
! 1.1-0   19-Dec-2003   Generate only one set of CR/CR/LF before ETX
!                       instead of two.           D. Offiler 
! 1.2-0   10-Jul-2004   Generate only 8 zeros for dummy (EOF) message
!                       instead of 10.            D. Offiler
!
! End of header ---------------------------------------------------------------
!
  IMPLICIT NONE

! Fixed values

  CHARACTER (LEN=1), PARAMETER :: zero      = "0"
  INTEGER,           PARAMETER :: Nend      = 4
  INTEGER,           PARAMETER :: end(Nend) = (/ 13, 13, 10, 3 /)
  INTEGER,           PARAMETER :: nul       = 0
  INTEGER,           PARAMETER :: LenDum    = 8

! Argument list parameters

  CHARACTER (LEN=*), INTENT(INOUT) :: Message    ! GTS message string
  INTEGER,           INTENT(IN)    :: GTShdrType ! GTS header code 
  INTEGER,           INTENT(INOUT) :: LenMsg     ! Pointer in Message

! Local parameters

  INTEGER :: lenm
  INTEGER :: ib

  IF ( GTShdrType >= 1 .AND. &
       GTShdrType <= 3 ) THEN

    lenm = LEN ( Message )

! For IP headers, but message < IP header length, just generate
! a dummy IP header

    IF ( GTShdrType == 3 .AND. &
         LenMsg     <= 11 ) THEN
      Message(1:LenDum) = REPEAT(zero,LenDum)
      LenMsg = LenDum

! If message string not long enough for trailer, 
! return length as zero bytes
 
    ELSE IF ( (LenMsg+Nend-1) > lenm ) THEN
      LenMsg = 0

! Normal trailer generation; set remaining message string to nulls

    ELSE
      DO  ib = 1, Nend
        Message(LenMsg:LenMsg) = CHAR(end(ib))
        LenMsg = LenMsg + 1
      END DO
      
      IF ( LenMsg < lenm ) &
        Message(LenMsg:lenm) = REPEAT(CHAR(nul),lenm-LenMsg+1)
      
      LenMsg = LenMsg - 1
      
! For IP headers, fill in pad 0's with actual length of message

      IF ( GTShdrType == 3 ) &
        WRITE ( Message(1:8), "(I8.8)") LenMsg - 10
      
    END IF

! No header/trailer - move pointer back to last character in Message.

  ELSE
    LenMsg = LenMsg - 1
  
  END IF
  
END SUBROUTINE GTSend
